{* Matrix Made Easy. Complex function                           *}
{* Author: Andrew N. Tuseyev                                    *}
{* Version 0.2b                                                 *}
{* Description: Complex matrix class and function of            *}
{*              complex variable                                *}
{* If you have some idea contact with me                        *}
{* E-Mail: brain@uti.ru                                         *}
{* All file provided "as is" without  warranties of stable work *}
{*--------------------------------------------------------------*}
{* If this library was useful to you send me some something by  *}
{* mail (like postcard, postmark, map of your city and etc)     *}
{* Thanx a lot.                                                 *}

unit MComplex;

interface

uses SysUtils,math;

type Tcomplex=record
      r:double;
      i:double;
     end;

function ComplexToStr(z:tcomplex):String;
function StrToComplex(z:string):Tcomplex;

// Complex Operation

function ComplexEqual(z1,z2:Tcomplex):Boolean;
function ComplexEqualZero(z:Tcomplex):Boolean;
function ComplexEqualOne(z:Tcomplex):Boolean;

function AddComplex(z1,z2:Tcomplex):Tcomplex;
function SubComplex(z1,z2:Tcomplex):Tcomplex;
function MulComplex(z1,z2:Tcomplex):Tcomplex;
function DivComplex(z1,z2:Tcomplex):Tcomplex;

function AddComplexArray(const Z: Array of TComplex):Tcomplex;
function SubComplexArray(const Z: Array of TComplex):Tcomplex;
function MulComplexArray(const Z: Array of TComplex):Tcomplex;
function DivComplexArray(const Z: Array of TComplex):Tcomplex;

function NegativeComplex(z:Tcomplex):Tcomplex;
function ModuleComplex(z:Tcomplex):double;
function InversComplex(z:Tcomplex):Tcomplex;
function NullComplex:Tcomplex;
function MakeComplex(z1,z2:double):Tcomplex;

function PolarComplex(z1,z2:double):Tcomplex;

function ArgumentComplex(z1:Tcomplex):double;
function ConjugeComplex(z1:Tcomplex):Tcomplex; // z:=x+iy result z*:=x-iy

function MulComplexParI(z1:Tcomplex):Tcomplex;
function MulComplexParNegI(z1:Tcomplex):Tcomplex;

function ComplexEXP(z:Tcomplex):Tcomplex;
function ComplexLN(z:Tcomplex):Tcomplex;
function ComplexSQRT(z:Tcomplex):TComplex;

function ComplexCOS(z:Tcomplex):Tcomplex;
function ComplexSIN(z:Tcomplex):Tcomplex;
function ComplexTAN(z:TComplex):Tcomplex;
function ComplexCOTAN(z:TComplex):Tcomplex;
function ComplexSEC(z:TComplex):Tcomplex;
function ComplexCOSEC(z:TComplex):Tcomplex;

function ComplexARCCOS(z:Tcomplex):TComplex;
function ComplexARCSIN(z:Tcomplex):TComplex;
function ComplexARCTAN(z:Tcomplex):TComplex;

function ComplexARGCOSH(z:Tcomplex):Tcomplex; // argch(z)
function ComplexARGSINH(z:Tcomplex):Tcomplex; // argsh(z)
function ComplexARGTANH(z:Tcomplex):Tcomplex; // argth(z)

function ComplexCOSH(z:Tcomplex):Tcomplex; // Cosh(z)
function ComplexSINH(z:Tcomplex):Tcomplex; // Sinh(z)
function ComplexTANH(z:Tcomplex):Tcomplex; // Tanh(z)
function ComplexCOTANH(z:TComplex):Tcomplex;
function ComplexSECH(z:TComplex):Tcomplex;
function ComplexCOSECH(z:TComplex):Tcomplex;


function TakeRealPart(z:TComplex):double;
function TakeImagePart(z:TComplex):double;

implementation

function TakeRealPart(z:TComplex):double;
begin
 Result:=Z.r;
end;

function TakeImagePart(z:TComplex):double;
begin
 Result:=Z.i;
end;


function ComplexToStr(z:tcomplex):String;
begin
if z.r<>0 then result:=FloatTostr(z.r) else result:='';
if (z.i>0) and (z.r<>0) then result:=result+'+';
if z.i<>0 then result:=result+FloatTostr(z.i)+'i' else if result='' then result:='0';
end;

function StrToComplex(z:string):Tcomplex;
var i:integer;
    sr,si:string;
begin
if z[Length(z)]in ['i','I'] then
  begin
  i:=Length(z)-1;
  while (not (z[i] in ['+','-'])) and (i>1) do dec(i);
  if z[i-1] in ['E','e'] then
     begin
     dec(i);
     while not (z[i] in ['+','-']) do dec(i);
     end;
  sr:=copy(z,1,i-1); if sr='' then sr:='0';
  si:=copy(z,i,length(z)-i);
  result.r:=StrToFLoat(sr);
  result.i:=StrToFLoat(si);
  end
else
  begin
  result.r:=StrTOFloat(z);
  result.i:=0;
  end;
end;

// Working With Complex Variables

function ComplexEqual(z1,z2:Tcomplex):Boolean;
begin
result:=(z1.r = z2.r) and (z1.i = z2.i);
end;

function ComplexEqualZero(z:Tcomplex):Boolean;
begin
result:=( z.r = 0 ) and ( z.i = 0 );
end;

function ComplexEqualOne(z:Tcomplex):Boolean;
begin
result:=( z.r = 1 ) and ( z.i = 0 );
end;


function AddComplex(z1,z2:Tcomplex):Tcomplex;// z1+z2
begin
result.r:=z1.r+z2.r;
result.i:=z1.i+z2.i;
end;

function SubComplex(z1,z2:Tcomplex):Tcomplex;// z1-z2
begin
result.r:=z1.r-z2.r;
result.i:=z1.i-z2.i;
end;

function MulComplex(z1,z2:Tcomplex):Tcomplex;// z1*z2
begin
result.r:=z1.r*z2.r-z1.i*z2.i;
result.i:=z1.r*z2.i+z1.i*z2.r;
end;

function DivComplex(z1,z2:Tcomplex):Tcomplex;// z1/z2
var descrim:double;
begin
 descrim:=z2.r*z2.r+z2.i*z2.i;
 if descrim<>0 then
   begin
   result.r:=(z1.r*z2.r+z1.i*z2.i)/descrim;
   result.i:=(z1.i*z2.r-z1.r*z2.i)/descrim;
   end
 else
   begin
   raise EDivByZero.CreateFmt('Evaluting Can''t Complete Because Descriminant of %s Equal Zero',[ComplexToStr(z2)]);
   result:=NullComplex;
   end;
end;

function AddComplexArray(const Z: Array of TComplex):Tcomplex;
var i:integer;
begin
result:=NullComplex;
for I := 0 to High(Z) do Result:=AddComplex(Result,z[i]);
end;

function SubComplexArray(const Z: Array of TComplex):Tcomplex;
var i:integer;
begin
result:=NullComplex;
for I := 0 to High(Z) do Result:=SubComplex(Result,z[i]);
end;


function MulComplexArray(const Z: Array of TComplex):Tcomplex;
var i:integer;
begin
result:=MakeComplex(1,0);
for I := 0 to High(Z) do Result:=MulComplex(Result,z[i]);
end;

function DivComplexArray(const Z: Array of TComplex):Tcomplex;
var i:integer;
begin
result:=z[0];
for I := 1 to High(Z) do Result:=AddComplex(Result,z[i]);
end;


function NegativeComplex(z:Tcomplex):Tcomplex; // -z
begin
  result.r:=-z.r;
  result.i:=-z.i;
end;

function ModuleComplex(z:Tcomplex):double; // |z|
begin
  result:=sqrt(z.r*z.r+z.i*z.i);
end;

function InversComplex(z:Tcomplex):Tcomplex; // 1/z
var descrim:double;
begin
descrim:=z.r*z.r+z.i*z.i;
 if descrim<>0 then
   begin
   result.r:=z.r/descrim;
   result.i:=-z.i/descrim;
   end
 else
   begin
   raise EDivByZero.CreateFmt('Evaluting Can''t Complete Because Descriminant of %s Equal Zero',[ComplexToStr(z)]);
   result:=NullComplex;
   end;
end;

function NullComplex:Tcomplex;
begin
 result.r:=0;
 result.i:=0;
end;

function MakeComplex(z1,z2:double):Tcomplex;
begin
 result.r:=z1;
 result.i:=z2;
end;

function ArgumentComplex(z1:Tcomplex):double; //  
begin
 result:=arctan2(z1.i,z1.r);
end;

function ConjugeComplex(z1:Tcomplex):Tcomplex; //  
begin
 result.r:=z1.r;
 result.i:=-z1.i;
end;

function MulComplexParI(z1:Tcomplex):Tcomplex; // z*i
begin {z=x+y*i -> -y+x*i}
result.r:=-z1.i;
result.i:=z1.r;
end;

function MulComplexParNegI(z1:Tcomplex):Tcomplex; // z*(-i)
begin {z=x+y*i -> y-x*i}
result.r:=z1.i;
result.i:=-z1.r;
end;

function ComplexEXP(z:Tcomplex):Tcomplex; // exp(z)
var temp:double;
begin {e xp(x+y*i)=exp(x)*(cos(y)+i*sin(y)) }
 temp:=exp(z.r);// for Speed UP
 result.r:=temp*cos(z.i);
 result.i:=temp*sin(z.i);
end;

function ComplexLN(z:Tcomplex):Tcomplex;
var Temp:double;
begin
temp:=z.r*z.r+z.i*z.i;
if temp=0 then
  begin
  raise EMathError.CreateFmt('Evaluting Can''t Complete Because Descriminant of %s Equal Zero',[ComplexToStr(z)]);
  //  result.error:=3; // Zero Logarithme raise
  result:=NullComplex;
  end
else
  begin
  result.r:=ln(temp);
  result.i:=ArgumentComplex(z);
  end;
end;

function ComplexSQRT(z:Tcomplex):TComplex; // sqrt(z)
var temp1,temp2:double;
begin
if (z.r<>0) or (z.i<>0) then
 begin
 temp1:=sqrt(0.5*(abs(z.r)+ModuleComplex(z)));
 temp2:=z.i/(2*temp1);
 if z.r>=0 then with result do
    begin
    r:=temp1;
    i:=temp2;
    end
 else if z.i<0 then with result do
    begin
    r:=-temp2;
    i:=-temp1;
    end
 else with result do
    begin
    r:=temp2;
    i:=temp1;
    end;
 end
else result:=z;
end;

function ComplexSIN(z:Tcomplex):Tcomplex;
begin
 { sin(x+i*y)=sin(x)*cos(i*y)+cos(x)*sin(i*y) sin(i*y)=i*sh(y) cos(i*x)=ch(x) }
 result.r:=sin(z.r)*cosh(z.i);
 result.i:=cos(z.r)*sinh(z.i);
end;

function ComplexCOS(z:Tcomplex):Tcomplex;
begin
 { cos(x+y*i)=cos(x)*cos(i*y)-sin(x)*sin(i*y) Look Up Man!!!}
 result.r:=cos(z.r)*cosh(z.i);
 result.i:=-sin(z.r)*sinh(z.i);
end;

function ComplexTAN(z:TComplex):Tcomplex;
var T1,t2:Tcomplex;
begin
t1:=ComplexCOS(z);
if ComplexEqualZero(t1) then
   begin
   raise EMathError.CreateFmt('Evaluting Can''t Complete Because Complex Cosinus of %s Equal Zero',[ComplexToStr(z)]);
   result:=NullComplex;
   end
else
   begin
   t2:=ComplexSIN(z);
   result:=DivComplex(t2,t1);
   end;
end;

function ComplexCOTAN(z:TComplex):Tcomplex;
begin
result:=InversComplex(ComplexTAN(z));
end;

function ComplexSEC(z:TComplex):Tcomplex;
begin
result:=InversComplex(ComplexCOS(z));
end;

function ComplexCOSEC(z:TComplex):Tcomplex;
begin
result:=InversComplex(ComplexSIN(z));
end;

function ComplexARGCOSH(z:Tcomplex):Tcomplex; // argch(z);
var t:tcomplex;
begin {argch(z)=-ln(z+i*sqrt(1-z*z)) }
t:=SubComplex(MakeComplex(1,0),MulComplex(z,z));
{t.r:=1-z.r*z.r+z.i*z.i;
t.i:=-2*z.r*z.i;}
result:=NegativeComplex(ComplexLN(AddComplex(z,MulComplexParI(ComplexSQRT(t)))));
end;

function ComplexARGSINH(z:Tcomplex):Tcomplex; // argch(z);
var t:tcomplex;
begin {argsh(z)=ln(z+i*sqrt(1+z*z)) }
t:=AddComplex(MakeComplex(1,0),MulComplex(z,z));
{t.r:=1+z.r*z.r-z.i*z.i;
t.i:=2*z.r*z.i;}
result:=ComplexLN(AddComplex(z,ComplexSQRT(t)));
end;

function ComplexARGTANH(z:Tcomplex):Tcomplex; // argth(z)
begin {argth(z)=(1/2)*ln((z+1)/(1-z)) }
if ComplexEqualZero(SubComplex(MakeComplex(1,0),z)) then
   begin
   raise EDivByZero.Create('Evaluting Can''t Complete Because Dividing By Zero');
   result:=NullComplex;
   end
else
   Result:=DivComplex(ComplexLN(DivComplex(AddComplex(z,MakeComplex(1,0)),SubComplex(MakeComplex(1,0),z))),MakeComplex(2,0));
end;

function ComplexCOSH(z:Tcomplex):Tcomplex; // Cosh(z)
begin { ch(x+i*y)=ch(x)*ch(i*y)+sh(x)*sh(i*y) }
result.r:=cosh(z.r)*cos(z.i);
result.i:=sinh(z.r)*sin(z.i);
end;

function ComplexSINH(z:Tcomplex):Tcomplex; // Sinh(z)
begin { sh(x+i*y)=sh(x)*ch(i*y)+ch(x)*sh(i*y) }
result.r:=sinh(z.r)*cos(z.i);
result.i:=cosh(z.r)*sin(z.i);
end;

function ComplexTANH(z:Tcomplex):Tcomplex; // Tanh(z)
begin
result:=DivComplex(ComplexSINH(z),ComplexCOSH(z))
end;

function ComplexCOTANH(z:TComplex):Tcomplex;
begin
result:=InversComplex(ComplexTANH(z));
end;

function ComplexSECH(z:TComplex):Tcomplex;
begin
result:=InversComplex(ComplexCOSH(z));
end;

function ComplexCOSECH(z:TComplex):Tcomplex;
begin
result:=InversComplex(ComplexSINH(z));
end;

function ComplexARCCOS(z:Tcomplex):TComplex;
begin { arccos(z)=-i*argch(z) }
result:=MulComplexParNegI(ComplexARGCOSH(z));
end;

function ComplexARCSIN(z:Tcomplex):TComplex;
begin { arcsin(z)=-i*argsh(i*z) }
result:=MulComplexParNegI(ComplexARGSINH(MulComplexParI(z)));
end;

function ComplexARCTAN(z:Tcomplex):TComplex;
begin { arctg(z)=-i*argth(i*z) }
result:=MulComplexParNegI(ComplexARGTANH(MulComplexParI(z)));
end;

function PolarComplex(z1,z2:double):Tcomplex;
begin
result.r:=z1*cos(z2);
result.i:=z1*sin(z2);
end;


end.
